package records

import (
	"encoding/binary"
	"errors"
	"strconv"

	"git.frostfs.info/TrueCloudLab/frostfs-node/cmd/frostfs-lens/internal/schema/common"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/shard"
	cid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/container/id"
	objectSDK "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object"
	oid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object/id"
)

var (
	ErrInvalidKeyLength   = errors.New("invalid key length")
	ErrInvalidValueLength = errors.New("invalid value length")
	ErrInvalidPrefix      = errors.New("invalid prefix")
)

func GraveyardRecordParser(key, value []byte) (common.SchemaEntry, common.Parser, error) {
	if len(key) != 64 {
		return nil, nil, ErrInvalidKeyLength
	}
	if len(value) != 64 {
		return nil, nil, ErrInvalidValueLength
	}
	var (
		cnr cid.ID
		obj oid.ID
		r   GraveyardRecord
	)

	_ = cnr.Decode(key[:32])
	_ = obj.Decode(key[32:])

	r.object.SetContainer(cnr)
	r.object.SetObject(obj)

	_ = cnr.Decode(value[:32])
	_ = obj.Decode(value[32:])

	r.tombstone.SetContainer(cnr)
	r.tombstone.SetObject(obj)

	return &r, nil, nil
}

func GarbageRecordParser(key, _ []byte) (common.SchemaEntry, common.Parser, error) {
	if len(key) != 64 {
		return nil, nil, ErrInvalidKeyLength
	}
	var (
		cnr cid.ID
		obj oid.ID
		r   GarbageRecord
	)

	_ = cnr.Decode(key[:32])
	_ = obj.Decode(key[32:])

	r.addr.SetContainer(cnr)
	r.addr.SetObject(obj)

	return &r, nil, nil
}

func ContainerVolumeRecordParser(key, value []byte) (common.SchemaEntry, common.Parser, error) {
	if len(key) != 32 {
		return nil, nil, ErrInvalidKeyLength
	}
	if len(value) != 8 {
		return nil, nil, ErrInvalidValueLength
	}
	var r ContainerVolumeRecord

	_ = r.id.Decode(key)
	r.volume = binary.LittleEndian.Uint64(value)

	return &r, nil, nil
}

func LockedRecordParser(key, value []byte) (common.SchemaEntry, common.Parser, error) {
	var (
		r   LockedRecord
		err error
	)

	if err := r.id.Decode(key); err != nil {
		return nil, nil, err
	}
	if r.ids, err = DecodeOIDs(value); err != nil {
		return nil, nil, err
	}
	return &r, nil, nil
}

func ShardInfoRecordParser(key, value []byte) (common.SchemaEntry, common.Parser, error) {
	if len(key) == 0 {
		return nil, nil, ErrInvalidKeyLength
	}

	var r ShardInfoRecord
	if string(key) == "id" {
		r.label = string(key)
		r.value = shard.ID(value).String()

		return &r, nil, nil
	}

	if len(value) != 8 {
		return nil, nil, ErrInvalidValueLength
	}
	r.label = string(key)
	r.value = strconv.FormatUint(binary.LittleEndian.Uint64(value), 10)

	return &r, nil, nil
}

func ObjectRecordParser(key, value []byte) (common.SchemaEntry, common.Parser, error) {
	if len(key) != 32 {
		return nil, nil, ErrInvalidKeyLength
	}
	var r ObjectRecord

	_ = r.id.Decode(key)
	if err := r.object.Unmarshal(value); err != nil {
		return nil, nil, err
	}

	return &r, nil, nil
}

func SmallRecordParser(key, value []byte) (common.SchemaEntry, common.Parser, error) {
	var r SmallRecord
	if err := r.id.Decode(key); err != nil {
		return nil, nil, err
	}
	if len(value) != 0 {
		x := string(value)
		r.storageID = &x
	}
	return &r, nil, nil
}

func RootRecordParser(key, value []byte) (common.SchemaEntry, common.Parser, error) {
	var r RootRecord
	if err := r.id.Decode(key); err != nil {
		return nil, nil, err
	}
	if len(value) == 0 {
		return &r, nil, nil
	}
	r.info = &objectSDK.SplitInfo{}
	if err := r.info.Unmarshal(value); err != nil {
		return nil, nil, err
	}
	return &r, nil, nil
}

func OwnerRecordParser(key, _ []byte) (common.SchemaEntry, common.Parser, error) {
	var r OwnerRecord
	if err := r.id.Decode(key); err != nil {
		return nil, nil, err
	}
	return &r, nil, nil
}

func UserAttributeRecordParser(key, _ []byte) (common.SchemaEntry, common.Parser, error) {
	var r UserAttributeRecord
	if err := r.id.Decode(key); err != nil {
		return nil, nil, err
	}
	return &r, nil, nil
}

func PayloadHashRecordParser(key, value []byte) (common.SchemaEntry, common.Parser, error) {
	if len(key) != 32 {
		return nil, nil, ErrInvalidKeyLength
	}
	var (
		err error
		r   PayloadHashRecord
	)

	r.checksum.SetSHA256([32]byte(key))
	if r.ids, err = DecodeOIDs(value); err != nil {
		return nil, nil, err
	}
	return &r, nil, nil
}

func ParentRecordParser(key, value []byte) (common.SchemaEntry, common.Parser, error) {
	var (
		r   ParentRecord
		err error
	)
	if err = r.parent.Decode(key); err != nil {
		return nil, nil, err
	}
	if r.ids, err = DecodeOIDs(value); err != nil {
		return nil, nil, err
	}
	return &r, nil, nil
}

func SplitRecordParser(key, value []byte) (common.SchemaEntry, common.Parser, error) {
	var (
		err error
		r   SplitRecord
	)
	if err = r.id.UnmarshalBinary(key); err != nil {
		return nil, nil, err
	}
	if r.ids, err = DecodeOIDs(value); err != nil {
		return nil, nil, err
	}
	return &r, nil, nil
}

func ContainerCountersRecordParser(key, value []byte) (common.SchemaEntry, common.Parser, error) {
	if len(value) != 24 {
		return nil, nil, ErrInvalidValueLength
	}

	var r ContainerCountersRecord
	if err := r.id.Decode(key); err != nil {
		return nil, nil, err
	}

	r.logical = binary.LittleEndian.Uint64(value[:8])
	r.physical = binary.LittleEndian.Uint64(value[8:16])
	r.user = binary.LittleEndian.Uint64(value[16:24])

	return &r, nil, nil
}

func ECInfoRecordParser(key, value []byte) (common.SchemaEntry, common.Parser, error) {
	var (
		r   ECInfoRecord
		err error
	)

	if err := r.id.Decode(key); err != nil {
		return nil, nil, err
	}
	if r.ids, err = DecodeOIDs(value); err != nil {
		return nil, nil, err
	}
	return &r, nil, nil
}
