package tracing

import (
	"crypto/x509"
	"errors"
	"fmt"
	"os"
	"strconv"

	"git.frostfs.info/TrueCloudLab/frostfs-node/cmd/frostfs-node/config"
	"git.frostfs.info/TrueCloudLab/frostfs-node/misc"
	"git.frostfs.info/TrueCloudLab/frostfs-observability/tracing"
)

const (
	subsection = "tracing"
)

// ToTracingConfig extracts tracing config.
func ToTracingConfig(c *config.Config) (*tracing.Config, error) {
	conf := &tracing.Config{
		Enabled:    config.BoolSafe(c.Sub(subsection), "enabled"),
		Exporter:   tracing.Exporter(config.StringSafe(c.Sub(subsection), "exporter")),
		Endpoint:   config.StringSafe(c.Sub(subsection), "endpoint"),
		Service:    "frostfs-node",
		InstanceID: getInstanceIDOrDefault(c),
		Version:    misc.Version,
		Attributes: make(map[string]string),
	}

	if trustedCa := config.StringSafe(c.Sub(subsection), "trusted_ca"); trustedCa != "" {
		caBytes, err := os.ReadFile(trustedCa)
		if err != nil {
			return nil, fmt.Errorf("cannot read trusted ca cert by path: %w", err)
		}
		certPool := x509.NewCertPool()
		ok := certPool.AppendCertsFromPEM(caBytes)
		if !ok {
			return nil, errors.New("can't fill cert pool by ca cert")
		}
		conf.ServerCaCertPool = certPool
	}

	i := uint64(0)
	for ; ; i++ {
		si := strconv.FormatUint(i, 10)
		ac := c.Sub(subsection).Sub("attributes").Sub(si)
		k := config.StringSafe(ac, "key")
		if k == "" {
			break
		}
		v := config.StringSafe(ac, "value")
		if v == "" {
			return nil, fmt.Errorf("empty tracing attribute value for key %s", k)
		}
		if _, ok := conf.Attributes[k]; ok {
			return nil, fmt.Errorf("tracing attribute key %s defined more than once", k)
		}
		conf.Attributes[k] = v
	}

	return conf, nil
}

func getInstanceIDOrDefault(c *config.Config) string {
	s := config.StringSliceSafe(c.Sub("node"), "addresses")
	if len(s) > 0 {
		return s[0]
	}
	return ""
}
